package Boggle::Round;

use strict;
use warnings;

use Boggle::Board;

use My::DBI qw(:all);
use DBI;
use DBD::Pg;
use Data::Dumper;


#States:
#a round is started and going on.
use constant B_STARTING => 1;
use constant B_RUNNING => 2;

sub new
{
   my $self = shift;
   my $args = shift;

   $args ||= {};

   my %defaults = (
      timeout => 180,
   );

   foreach my $k (keys %defaults) {
      $args->{config}->{$k} ||= $defaults{$k};
   }
   
   my $r = bless $args, $self;
   
   #TODO - configurability.
   $r->core->dbi_connect({ name => 'quiz', database => 'alex', username => 'alex'});
   $r->dbh($r->core->dbh('quiz'));

   return $r;
}

sub state 
{
   my $self = shift;
   my $state = shift;

   return $self->{state} unless $state;

   if($state == Q_QUESTION) {
      $self->{timeout} = time() + $self->config->{timeout};
      $self->core->removeAllTicks;
      $self->core->addTick($self->{timeout}, 'Quiz');
   } elsif($state == Q_PAUSE) {
      $self->core->removeAllTicks;
      $self->core->addTick(time() + $self->config->{pause}, 'Quiz');
   } elsif($state == Q_RPAUSE) {
      $self->core->removeAllTicks;
      $self->core->addTick(3600, 'Quiz'); #To end the round after some period of time. XXX - Configurability.
      delete $self->core->plugin->{public}->{$self->where};
   } elsif($state == Q_ENDED) {
      $self->core->removeAllTicks;
   }

   return $self->{state} = $state;
}

sub start
{
   my $self = shift;

   return if($self->{id}); #Already started.

   $self->{id} = dbi_exec_for_scalar($self->dbh, "SELECT nextval('rounds_id_seq')");

   dbi_save_from_hash($self->dbh, 'rounds','id', { id => $self->{id}, mode => $self->{mode} });

   $self->state(Q_PAUSE);

   #Take control. :)
   $self->core->plugin->{public}->{$self->where} = 'Quiz';
}

sub stop
{
   my $self = shift;
   delete $self->core->plugin->{public}->{$self->where};
   $self->state(Q_ENDED);
   delete $self->core->plugin->{public}->{$self->where};
}


=item abort

Aborts the round.

=cut

sub abort
{
   my $self = shift;

   return unless($self->{id}); #Not even started.

   $self->stop;

   prepare_and_exec($self->dbh, 'DELETE FROM rounds WHERE id=?', $self->id);
   $self->{id} = undef;
}

=item pause

Pauses the round.

=cut

sub pause
{
   my $self = shift;

   $self->state(Q_RPAUSE);
   $self->msg("Ingen svarer. Runde pauset. Skriv \"!q start\" for å starte igjen");
}

=item continue

Continues the round.

=cut

sub continue
{
   my $self = shift;

   $self->msg(sprintf("Fortsetter runde #%i", $self->id));

   $self->askNextQuestion;
}

=item id

Returns the round id. 

The round id is only available after the round is started.

=cut

sub id { return shift->{id}; }

=item mode

Returns the round mode.

B<0> is single player.

B<1> is team play.

=cut

sub mode { return shift->{mode}; }

=item askedQuestion ($Winner)

B<Winner> is a C<Quiz::User>-object of the user answering correctly -- which, naturally, may be undef. :-)

=cut

sub askedQuestion
{
   my $self = shift;
   my $Winner = shift;

   return unless($self->id); #Not started.

   return unless(ref($self->question));

   my %row;
   $row{round} = $self->id;
   $row{question} = $self->question->id;
   $row{winner} = $Winner->id if($Winner);

   dbi_insert_from_hash($self->dbh, 'round_questions', \%row);
}

=item getUser ($hostmask)

Returns a C<Quiz::User>-object, assuming it is participating.

=cut

sub getUser
{
   my $self = shift;
   my $hostmask = shift;

   if( $self->{users}->{$hostmask} ) {
      return $self->{users}->{$hostmask}; 
   } else {
      my $User = Quiz::User->GetByHostmask($self->dbh, $hostmask);
      unless(dbi_exists($self->dbh, 'select user_id from round_participants where round=? and user_id=?', $self->id, $User->id)) { # Workaround for weird bug.
	 dbi_insert_from_hash($self->dbh, 'round_participants', {
	       round => $self->id,
	       user_id => $User->id,
	 });
      }
      return $self->{users}->{$hostmask} = $User;
   }
}

=item moveUser ($hostmask, $newnick) 

Updates the stored hostmask for a user.

The user object must be retrievable even if the user changes its nick.

=cut

sub moveUser
{
   my $self = shift;
   my ($hostmask, $newnick) = @_;

   return 0 unless($self->{users}->{$hostmask});

   $hostmask =~ /!(.*)/;
   my $newhostmask = "$newnick!$1";
   $self->{users}->{$newhostmask} = $self->{users}->{$hostmask};
   delete $self->{users}->{$hostmask};

#   prepare_and_exec($self->dbh, 'UPDATE users SET hostmask=? WHERE hostmask=?', $newhostmask, $hostmask);

   return 1;

}

=item end ($Winner)

B<Winner> is a C<Quiz::User>-object of the user winning the round, if any.

=cut

sub end
{
   my $self = shift;
   my $Winner = shift;


   if($Winner) {
      my %oldstats = dbi_exec_for_row_hash($self->dbh, "SELECT *, 100 * answered::float/attempts::float AS ratio FROM rank_history WHERE id = ? ORDER BY date DESC LIMIT 1", $Winner->id);
      my $msg;

      prepare_and_exec($self->dbh, "UPDATE rounds SET winner=?, ended=now() WHERE id=?", $Winner->id, $self->id);
      my %newstats = %{ $Winner->stats };

      if($oldstats{total_score}) {
	 #End round, which will update the ranks.

	 #Some statistics:
	 my ($rank, $score, $ratio);
	 if($newstats{rank} > $oldstats{rank}) {
	    $rank = sprintf("#%i (-%i)", $newstats{rank}, $newstats{rank} - $oldstats{rank});
	 } elsif($newstats{rank} < $oldstats{rank}) {
	    $rank = sprintf("#%i (+%i)", $newstats{rank}, $oldstats{rank} - $newstats{rank});
	 } else {
	    $rank = "#$newstats{rank}";
	 }

	 #In case we ever add penalties:
	 if($newstats{score} > $oldstats{total_score}) {
	    $score = sprintf("%.2f (+%.2f)", $newstats{score}, $newstats{score} - $oldstats{total_score});
	 } elsif($newstats{score} < $oldstats{total_score}) {
	    $score = sprintf("%.2f (-%.2f)", $newstats{score}, $oldstats{total_score} - $newstats{score});
	 } else {
	    $score = sprintf("%.2f", $newstats{score});
	 }

	 if($newstats{ratio} > $oldstats{ratio}) {
	    $ratio = sprintf("%.2f%% (+%.2f%%)", $newstats{ratio}, $newstats{ratio} - $oldstats{ratio});
	 } elsif($newstats{ratio} < $oldstats{ratio}) {
	    $ratio = sprintf("%.2f%% (-%.2f%%)", $newstats{ratio}, $oldstats{ratio} - $newstats{ratio});
	 } else {
	    $ratio = sprintf("%.2f%%", $newstats{ratio});
	 }

	 $msg = sprintf("Runde #%i er ferdig. %s vant for %i. gang. %s sin statistikk: Rank: %s. Poeng: %s. Ratio: %s.",
		   $self->id, $Winner->nick, $newstats{wins}, $Winner->nick, $rank, $score, $ratio);
      } else {
	 $msg = sprintf("Runde #%i er ferdig. %s vant for %i. gang.", $self->id, $Winner->nick, $newstats{wins});
      }
      $self->core->isInfo;
      $self->core->msg($self->where, BOLD . $msg);


   } else {
      prepare_and_exec($self->dbh, "UPDATE rounds SET ended=now() WHERE id=?", $self->id);

      my $msg;
      if($self->state == Q_RPAUSE) {
	 # XXX - Configurability.
         $msg = sprintf("Runde #%i har vært pauset i en time. Avslutter runden. Ingen vinner.", $self->id);
      } else {
        $msg = sprintf("Runde #%i avsluttet. Ingen vinner.", $self->id);
     }
      $self->core->isInfo;
      $self->core->msg($self->where, BOLD . $msg);
   }

   my %rstat = dbi_exec_for_row_hash($self->dbh, "SELECT sum(attempts) AS attempts, sum(wins) AS wins FROM round_participants WHERE round=?", $self->id);
   my $qcount = dbi_exec_for_scalar($self->dbh, "SELECT count(*) FROM round_questions WHERE round=?", $self->id);

   my $percentage;
   if($qcount) {
      $percentage = 100 * $rstat{wins}/$qcount;
   } else { 
      $percentage = 0;
   }

   $self->core->msg($self->where, sprintf("Det ble stilt %i spørsmål hvorav det ble svart riktig på %i (%.2f%%) av dem. Totalt ble det gitt %i svar.",
	     $qcount, $rstat{wins}, $percentage, $rstat{attempts}));

   $self->state(Q_ENDED);
   delete $self->core->plugin->{public}->{$self->where};
}

=item askNextQuestion

Ask the next question.

=cut

sub askNextQuestion
{
   my $self = shift;
   $self->setTimeout;
   my $q = $self->nextQuestion;
   my $msg = sprintf("[%s] %s (%.2f poeng, #%i)", $q->category->breadcrumb, $q->question, $q->score, $q->id);
   $self->core->msg($self->where, BOLD . $msg);
   $self->question($q);
   $self->state(Q_QUESTION);
}

=item doTimeout

No correct answers and time is up.

=cut

sub doTimeout
{
   my $self = shift;
   $self->askedQuestion;
   $self->core->isInfo;
   $self->core->msg($self->where, BOLD . "Time's up, osv");
   $self->state(Q_PAUSE);
   #Save question with no winner.
   #Set state. Ask next question.
}

=item addParticipant ({ user => $u, attempts => x, wins => y })

B<user> is a C<Quiz::User>-object of the participating user.

B<attempts> is the number of times he/she has attempted to answer.

B<wins> is the number of times he/she has answered correctly. 

=cut

sub addParticipant
{
   my $self = shift;
   my $args = shift;

   return unless($self->{id}); #Not even started.
   return unless(ref($args->{user}) eq ('Quiz::User'));

   return if($args->{wins} > $args->{attempts});

   $args->{wins} ||= 0;

   prepare_and_exec("INSERT INTO round_participants(round, user_id, attempts, wins) VALUES (?,?,?)", $args->{round}->id, $args->{user}->id, $args->{attempts}, $args->{wins});

}

sub teamAddMember
{

}

=item setWinner ({ team/winner => $foo })

If the mode is B<0>, a C<Quiz::User>-object must be specified.

If the mode is B<1>, a C<Quiz::Team>-object must be specified.

=back

=cut

sub setWinner
{
   my $self = shift;
   my $args = shift;

   return unless($self->{id}); #Not even started.
   return unless(( !$self->mode and ref($args->{user}) eq 'Quiz::User') or ($self->mode == 1 and ref($args->{team}) eq 'Quiz::Team'));

   if($args->{team}) {
      return if(!dbi_exists($self->dbh, 'SELECT round FROM round_teams WHERE round=? AND team=?', $self->id, $self->{team}->id));

      dbi_save_from_hash($self->dbh, 'rounds', { winner_team => $self->{team}->id });
      return 1;
   } else {
      dbi_save_from_hash($self->dbh, 'rounds', { winner => $self->{user}->id });
      return 1;
   }
}

#TODO : ->end? Lagre metadata osv.

=item avgScore

Returns the average score of the questions eligible for the round.

=cut

sub avgScore
{
   my $self = shift;

   my $args = $self->{args};
   $args->{Select} = "AVG(question_score(asked,answered))";
   my $query = Quiz::Question->Query($args);

   return dbi_exec_for_scalar($self->dbh, $query);
}

=item config

Returns the round configuration.

=cut

sub config { return shift->{config}; }

=item config

Returns the core object.

=cut

sub core { return shift->config->{core}; }
sub where { return shift->config->{where}; }

sub msg
{
   my $self = shift;
   my $msg = shift;

   $self->core->isInfo;

   $self->core->msg($self->where, BOLD . $msg);
}

sub ping
{
   my $self = shift;
   $self->{act} = time();
}

sub act
{
   my $self = shift;
   return $self->{act};
}

sub channel { return shift->{Channel}; }

#get-ers and set-ers.
sub AUTOLOAD {
    my $self = shift;
    my $attr = our $AUTOLOAD;
    $attr =~ s/.*:://;
    return unless $attr =~ /[^A-Z]/;  # skip DESTROY and all-cap methods

    $self->{lc $attr} = shift if @_;
   
    return $self->{lc $attr};
}

=head1 SEE ALSO

L<Quiz::Question>
L<Quiz::User>
L<Quiz::Category>

=head1 AUTHOR

Alex Brasetvik 

=cut

1;
