package Quiz;

use lib '../';
use warnings;
use strict;

use vars qw($DBH $FP $core);

use My::DBI qw(:all);
use DBI;
use DBD::Pg;

use Quiz::Round;
use Quiz::Question;
use Quiz::User;
use Quiz::Category;

use Term::ANSIColor qw(:constants);

use Common;
use Locale::gettext;
use POSIX;
setlocale(LC_MESSAGES, "");
bindtextdomain('erwin', 'locale/');
textdomain('erwin');

use base qw(Plugin);

#Regex for matching all number representations. Cheers, Regexp::Common
$FP = '(?:(?i)(?:[+-]?)(?:(?=[0123456789]|[.])(?:[0123456789]*)(?:(?:[.])(?:[0123456789]{0,}))?)(?:(?:[E])(?:(?:[+-]?)(?:[0123456789]+))|))';

#States:
#a round is started and going on.
use constant Q_RUNNING => 1;

#a question is asked.
use constant Q_QUESTION => 2;

#pause between questions - aka. "sleep period".
use constant Q_PAUSE => 3;

#the round is paused.
use constant Q_RPAUSE => 4;

#the round is ended and may be overwritten.
use constant Q_ENDED => 5;

#gathering players for team auto generation.
use constant Q_GATHER => 6;

#Start a new round
sub Start
{
   my($core, $backrefs, $args) = @_;
   
   if($core->heap->{privmsg}) { #In case of a badly configured regex.
      $core->log('Badly configured regex! Quiz::Start enabled for private messages.', WARNING);
      return;
   }
   
   return unless(Quiz::User->GetByHostmask($core->dbh('quiz'), $core->heap->{who})); #Only logged in users are allowed to start.

   #Retrieve configurations.
   my %config;
   $core->module_config(\%config, $core->where);
   return unless $config{enabled}; #Not enabled for this channel.

   $config{core} = $core;
   $config{where} = $core->where;

   my $R = $core->heap->{Q}->{$core->where};

   if($R && $R->state == Q_RPAUSE) {
      if(time() - $R->act > 600) {
	 #No.
	 $R->end;
      } else {
         $R->continue;
      }
   }
   return if($R && $R->state != Q_ENDED); #A round is already started.

   #Figure out rules and create a new round. 
   if($args->{rules}) {
      #Rules are defined by the calling regex. E.g. "!sport_quiz" or "!easy".
      $args->{rules}->{config} ||= \%config;
      $R = Quiz::Round->new($args->{rules});
   } else {
      my $options = $backrefs->[0]; #eg. "+geography -asia >1.0"
      if(!$options) {
         $R = Quiz::Round->new({config => \%config});
      } else {
         my @in = $options =~ m/\+(\w+) /g;
	 my @notIn = $options =~ m/\-(\w+) /g;

         my(@invalid); #verify:
         foreach my $category (@in, @notIn) {
	    my $C;
	    if($category =~ /\D/) {
	       $C = Quiz::Category->GetByName($category);
	    } else {
	       $C = Quiz::Category->GetByID($category);
	    }
	    push(@invalid, $category) unless $C;
	 }

         if(@invalid) {
	    return replace(gettext('Invalid categories: $X'), { X => join ", ", @invalid });
	 }

	 # TODO - Team play
	 #my $tp = gettext('teamplay');
	 # TODO - Lightning round?
	 #my $mode = 1 if($options =~ /:$tp/i);
	 my $mode = 0;

	 my ($diffgt) = $options =~ m/>($FP)/;
	 my ($difflt) = $options =~ m/<($FP)/;

	 $difflt = undef if $diffgt;

	 my %args = (
	        Channel => $core->where,
	 	CategoryNotIn => \@notIn,
		CategoryIn => \@in,
		DiffLT => $diffgt,
		DiffGT => $diffgt,
		Mode => $mode,
		config => \%config
	 );
         $R = Quiz::Round->new(\%args);
      }
   }

   if($R->numberOfQuestions < $R->config->{minimum_questions}) {
      return replace(gettext('Too few questions with those requirements. ($X). Need at least $Y questions to start a round.'),
		    { X => $R->numberOfQuestions, Y => $config{minimum_questions} });
   }

   #TODO - generate teams, etc.
   $core->setMode($core->where, '+m');

   $core->heap->{Q}->{$core->where} = $R;
   $R->start;
   $R->ping;



   # TODO - start round sanely.
   $core->isInfo;
   return "AW RITE";
}

#Top(X) stats.
sub Top
{
  my($core, $backrefs, $args) = @_;
  return if($core->testing);

  my $n = $backrefs->[0];
  my $ratio = $n =~ s/ratio//g;

  $n =~ s/\s//g;
  $n = 3 if($n =~ /\D/);
  $n ||= 3;

  $n = 10 if($n > 10);

  my @stats;
  if($ratio) {
     @stats = dbi_exec_for_loh($core->dbh('quiz'), 'select users.username, score_cache.total_score, 100 * float4(answered)/float4(attempts) as ratio from score_cache, users where users.id=score_cache.id order by ratio desc limit ?', $n);
  } else {
     @stats = dbi_exec_for_loh($core->dbh('quiz'), 'SELECT rank, username, score FROM top(?)', $n);
  }

  for(my $x=0; $x <= $#stats; $x++) {
     if($ratio) {
        $stats[$x] = sprintf("#%i: %s (%.2f%%, %.2fp)", $x + 1, $stats[$x]->{username}, $stats[$x]->{ratio}, $stats[$x]->{total_score});
     } else {
        $stats[$x] = sprintf("#%i: %s (%.2f)", $stats[$x]->{rank}, $stats[$x]->{username}, $stats[$x]->{score});
     }
  }

  return join ", ", @stats;
}

sub Low
{
  my($core, $backrefs, $args) = @_;
  return if($core->testing);
  my $n = $backrefs->[0];
  $n =~ s/\s//g;
  $n ||= 3;
  return if($n =~ /\D/);

  $n = 10 if($n > 10);

  my @stats = dbi_exec_for_loh($core->dbh('quiz'), 'SELECT rank, username, score FROM low(?)', $n);

  for(my $x=0; $x <= $#stats; $x++) {
     $stats[$x] = sprintf("#%i: %s (%.2f)", $stats[$x]->{rank}, $stats[$x]->{username}, $stats[$x]->{score});
  }

  return join ", ", @stats;
}

sub AvgScore
{
   return dbi_exec_for_scalar($core->dbh('quiz'), "SELECT AVG(question_score(answered, asked)) FROM questions");
}

sub Stats
{
   my($core, $backrefs, $args) = @_;
   return if($core->testing);

   my $username = $backrefs->[0];
   $username =~ s/\s//g;

   my (%stats, $uid);
   if($username =~ /^\w{4,9}$/) { # Input is probably a username.
      if(dbi_exists($core->dbh('quiz'), "SELECT username FROM users WHERE username=?", lc($username))) {
            %stats = dbi_exec_for_row_hash($core->dbh('quiz'), "SELECT * FROM userinfo WHERE username=?", lc($username));
      } elsif($uid = dbi_exec_for_scalar($core->dbh('quiz'), "SELECT username FROM users WHERE hostmask LIKE '%$username!%'")) { 
            %stats = dbi_exec_for_row_hash($core->dbh('quiz'), "SELECT * FROM userinfo WHERE username=?", $uid);
      } else {
	    return undef;
      }
   } elsif($username =~ /^[\w\[\]\|{}_\-\`]{2,9}$/) { #Input is probably an IRC-nick.
      if($uid = dbi_exec_for_scalar($core->dbh('quiz'), "SELECT username FROM users WHERE hostmask LIKE '%$username!%'")) {
         %stats = dbi_exec_for_row_hash($core->dbh('quiz'), "SELECT * FROM userinfo WHERE username=?", $uid);
      } else {
	 return undef;
      }
   } else {
    %stats = dbi_exec_for_row_hash($core->dbh('quiz'), "SELECT * FROM userinfo WHERE hostmask=?", $core->heap->{who});
   }
   if(keys %stats) {
       return sprintf("Statistikk for «%s»: Rank: #%i, Poeng: %.2f, Ratio: %.2f%%, Antall seiere: %i, Gj.poeng: %.2f",
       		    $stats{username}, $stats{rank}, $stats{score}, $stats{ratio}, $stats{wins}, $stats{score}/$stats{answered});
   } else {
      return undef;
   }
}

sub Comment
{
   my($core, $backrefs, $args) = @_;
   return if($core->testing);

   my $User = Quiz::User->GetByHostmask($core->dbh('quiz'), $core->heap->{who});
   return gettext('You are not logged in.') unless($User);

   my $qid = $backrefs->[0];
   return gettext('Invalid input. The question id is an integer.') if($qid =~ /\D/);

   #my $Question = Question->GetByID($core->dbh('quiz'), $qid);
   my $Question = Quiz::Question->GetByID($qid);

   if($Question) {
      dbi_insert_from_hash($core->dbh('quiz'), 'question_reports', { author => $User->id, comment => $backrefs->[1], question => $qid });
      return gettext('Comment to question added.');
   } else {
      return gettext('No question with that ID exists.');
   }
}

sub cron
{
   my $core = shift;

   foreach my $R (values %{ $core->heap->{Q} }) {
      #For each round
#	    return;
      if($R->state == Q_PAUSE) {
	 print "\t PAUSE\n";
	 #Pause is done. Ask question.
	 $R->askNextQuestion;
      } elsif($R->state == Q_RPAUSE) {
	 print "\t RPAUSE\n";
	 $R->end;
	 $core->setMode($R->channel, '-m');
      } elsif($R->state == Q_QUESTION) {
	 printf "\t QUESTION, %i\n", $R->act;
	 #A question is asked.
	 if($R->act && time() - $R->act >= 120) {
	    print "\t QUESTION -> PAUSE\n";
	    $R->pause;
	    $core->setMode($R->channel, '-m');
	    return;
	 }
	 if($R->timeout <= time()) {
	    print "\t TIMEOUT\n";
	    #Time is up
	    $R->doTimeout;
	 }
      } else {
	 return;
      }
   }
}

#Run on every incoming message.
sub public
{
   my($core, $who, $where, $msg) = @_;

   return if($msg =~ /:/);

   $who =~ /(.*)!/;
   return if($1 eq $core->mynick);


   my $R = $core->heap->{Q}->{$where};
   return unless $R;

=cut
   my $stop_regex = gettext('^!q stopp?$');

   if($msg =~ /$stop_regex/i) {
      my $R = $core->heap->{Q}->{$core->where};
      $R->end;
      delete $core->heap->{Q}->{$core->where};
      return;
   }
=cut

   my $comment_regex = gettext('!q ?comment (\d+) (.+)');
   if($msg =~ /$comment_regex/) {
      $core->msg($where, $core->heap->{nick} . ": " . Quiz::Comment($core, [$1, $2], {}));
      return;
   }

   $R->ping;

   if($R->state == Q_GATHER) {
      # TODO - Add player to team.
   } elsif ($R->state == Q_QUESTION) {
      #A questions is asked, and the user wants to answer:
     
      my $User = $R->getUser($who);

      if($R->question->check_answer($msg)) {

         #my $User = Quiz::User->GetByHostmask($R->dbh, $who);

	 my $score = dbi_exec_for_scalar($core->dbh('quiz'), "SELECT reduced_question_score(?,?)", $R->question->id, $User->id);
	 prepare_and_exec($R->dbh, "UPDATE round_participants SET attempts = attempts + 1, wins = wins + 1 WHERE round=? AND user_id=?", $R->id, $User->id);
	 $R->askedQuestion($User);

         my $a = $R->question->format_answer;

	 my $TRESHOLD = 10; # XXX
	 my $count = dbi_exec_for_scalar($core->dbh('quiz'), "SELECT count(*) FROM round_questions WHERE round = ? AND winner = ?", $R->id, $User->id);

	 $R->msg(sprintf("Riktig svar: %s. (%s, %i/%i, +%.02fp)", $a, $User->nick, $count, $TRESHOLD, $score));

	 if( $count >= $TRESHOLD ) {
	    $R->end( $User );
            $core->setMode($core->where, '-m');
	 } else {
	    $R->state(Q_PAUSE);
	 }
      } else {
	 prepare_and_exec($R->dbh, "UPDATE round_participants SET attempts = attempts + 1 WHERE round=? AND user_id=?", $R->id, $User->id);
      }
   } else {
     return; #?
   }
}

sub nick
{
   my($core, $who, $newnick) = @_;

   $who =~ /(.*)!/;
   return if($1 eq $core->mynick);

   printf "*** In nick() in Quiz.pm: $who, $newnick\n";

   foreach my $where (keys %{ $core->heap->{Q} }) {
      my $R = $core->heap->{Q}->{$where};
      next unless $R;
      printf "*** Calling R->moveUser()\n";
      $R->moveUser($who, $newnick);
   }

   $who =~ /!(.*)/;
   my $newhostmask = "$newnick!$1";
   #XXX : Configurability.
   printf "*** Updating users-table.\n";
   prepare_and_exec($core->dbh('quiz'), 'UPDATE users SET hostmask=? WHERE hostmask=?', $newhostmask, $who);

}

sub join
{
   my($core, $who, $where) = @_;

   print "*** in join(); \n";

   my %config;
   $core->module_config(\%config, $where);
   
   print "*** enabled? \n";
   return unless $config{enabled};
   print "*** yes \n";

   my $User = Quiz::User->GetByHostmask($core->dbh('quiz'), $who);
   $who =~ /(.*)!/; my $nick = $1;
   
   if($User) {
      $core->setMode($where,'+v', $nick);
      $core->msg($nick, sprintf("Hei. Du er logget inn som \"%s\".", $User->username));
   } else {
      $core->msg($nick, "Hei. Du er ikke logget inn. For å spille må du ha en bruker og være logget inn. Skriv " . BOLD . "!q auth" . RESET .
	 " for å logge inn eller " . BOLD . "!q register" . RESET . " for å registrere en bruker.");
   }
}


sub first_run {
   my $core = shift;
   $core->heap->{Q} ||= {}; #Create namespace.
   my %config;
   $core->module_config(\%config, $core->where);
   $core->dbi_connect({ database => $config{database}, username => $config{username}, password => $config{password}, name => 'quiz' });

}

sub Channels {
   my $core = shift;
  
   my @channels;
   foreach my $channel ( @{ $core->config->{Channel} } ) {
      my %config;
      $core->module_config(\%config, $channel);
      if($config{enabled}) {
	 push(@channels, $channel);
      }
   }
   return @channels;
}

__PACKAGE__->install({
	   events => { public => 1, nick => 1, join => 1 },
});

1;
