package Quiz::Category;

use strict;
use warnings;

use vars qw($dbh $delimiter);

use Quiz::Question;

use My::DBI qw(:all);
use DBI;
use DBD::Pg;
use Data::Dumper;

$dbh = DBI->connect('DBI:Pg:dbname=alex', 'alex');
$delimiter = ' :: ';

=item1 NAME

Quiz::Category

=item1 SYNOPSIS

 $c = Quiz::Category->GetByName('Geography');
 $c = Quiz::Category->GetByID(666);

=item1 DESCRIPTION

A category contains questions and other categories, and might be the child of a parent category.

=item1 METHODS

The following methods are available:

=over 4

=item new ({ OPTIONS })

Valid options are:

B<name>: The name of the category.

B<parent>: The id of a parent category.

Note that the combination of name and parent id must be unique.

Returns a C<Quiz::Category>-object.

=cut

sub new
{
   my $self = shift;
   my $args = shift;

   return unless($args->{name});

   return unless(Quiz::Category->GetByID($args->{id}));

   return bless $args, $self;

}

=item GetByName ('Geography')

Returns a C<Quiz::Category>-object given its name.

Note that the combination of a category parent and -name has to be unique.

=cut

sub GetByName
{
   my $self = shift;
   my $name = shift;

   my %category = dbi_exec_for_row_hash($dbh, "SELECT *, breadcrumb(id, ?) FROM categories WHERE LOWER(name)=?", $delimiter, lc($name));

   return bless \%category, $self;
}

=item GetByID(666)

Returns a C<Quiz::Category>-object given its id.

=cut

sub GetByID
{
   my $self = shift;
   my $id = shift;

   my %category = dbi_exec_for_row_hash($dbh, "SELECT *, breadcrumb(id, ?) FROM categories WHERE id=?", $delimiter, $id);

   return bless \%category, $self;
}

=item name ('new name')

Returns the category's name.

=cut

sub name
{
   my $self = shift;
   my $name = shift;

   if($name) {
      if($self->{parent}) {
         #Must be unique!
	 return if(dbi_exists($dbh, "SELECT * FROM categories WHERE LOWER(name)=? AND parent=?", lc($name), $self->parent));
      } else {
	 $self->{name} = $name;
      }
   }

   return $self->{name};
}

=item id

Returns the category id.

Read only.

=cut

sub id { return shift->{id}; }

=item parent

Returns a C<Quiz::Category>-object for the parent category, if it exists.

=cut

sub parent
{
   my $self = shift;
   my $parent = shift;

   if($parent) { #Trying to set.
      #Must be a unique combination!
      return if(dbi_exists($dbh, "SELECT * FROM categories WHERE LOWER(name)=? AND parent=?", lc($self->name), $parent));
      #Must exist!
      return unless(dbi_exists($dbh, "SELECT * FROM categories WHERE parent=?", $parent));

      $self->{parent} = $parent;
      $self->{Parent} = Quiz::Category->GetByID($self->{parent});
   }
      
   $self->{Parent} = Quiz::Category->GetByID($self->{parent}) unless $self->{Parent};

   return $self->{Parent};
}

sub save {
   my $self = shift;

   my %row = map { $_ => $self->{$_} } qw(id name parent);

   if($row{id}) {
      dbi_save_from_hash($dbh, 'categories','id', \%row);
   } else {
      dbi_insert_from_hash($dbh, 'categories', \%row);
   }
}

=item breadcrumb

Returns a "breadcrumb" for the category, e.g. "Geography, Capitals, Europe".

Read only.

=cut

sub breadcrumb
{
   my $self = shift;

   $self->{breadcrumb} ||= 'Ukategorisert'; #XXX - gettext

   return $self->{breadcrumb};

}

=item numberOfQuestions

Returns the number of question in this category.

=cut

sub numberOfQuestions
{
   my $self = shift;

   return dbi_exec_for_scalar($dbh, "
   SELECT count(*)
   FROM questions
   WHERE category IN (
   	SELECT id
	FROM walk_tree(?)
   )", $self->id);
}

1;
