package Quiz::Question;
use warnings;
use strict;

use vars qw($dbh);

use Quiz::Question;
use Quiz::User;
use Quiz::Category;
use lib '../..';
use My::DBI qw(:all);
use DBI;
use DBD::Pg;

use Data::Dumper;

$dbh = DBI->connect('DBI:Pg:dbname=alex', 'alex');

=head1 NAME

Quiz::Question

=head1 SYNOPSIS

 my $q = Quiz::Question->GetById(4);
 $q = Quiz::Question->GetRandom({ CategoryNotIn => [], CategoryIn => [], DiffLT => [], DiffGT => [] });

=head1 DESCRIPTION

A question is, uh, a question that belongs in a category, an answer, an author, number of times it has been
asked, and number of times it has been answered correctly. It can be marked for review, inactive (rather than
deleted, as there's probably other records depending on it), and knows the last time it has been asked.

=cut

=head1 METHODS

The following methods are available:

=over 4

=item GetRandom ({ OPTIONS });

Valid options are:

B<CategoryNotIn> is an array reference of category IDs the question is not allowed to be in.

B<CategoryIn> is an array reference of category IDs the question must be in.

B<DiffLT> is a float point number the question's difficulty/score must be Lower Than.

B<DiffGT> is a float point number the question's difficulty/score must be Greater Than.

Returns a C<Quiz::Question>-object.

=cut

sub GetRandom
{
   my $self = shift;
   my $args = shift;

   my $query = Quiz::Question->Query($args);

   $query .= " ORDER BY RANDOM() LIMIT 1";

   my %q = dbi_exec_for_row_hash($dbh, $query);

   $q{_args} = $args;

   return bless \%q, $self;
}

=item GetByID (123)

Returns a C<Quiz::Question>-object, given its id.

=cut

sub GetByID
{
   my $self = shift;
   my $id = shift;

   my %q = dbi_exec_for_row_hash($dbh, "SELECT * FROM questions WHERE id=?", $id);

   return bless \%q, $self;
}

=item regexes ()

Returns two arrayrefs of regexes. The first array contains regexes that the answer
*must* match to be accepted, if any. The second array contains regexes where at least one
of them must match the answer, if any.

=cut

sub regexes {
   my $self = shift;
   my $a = $self->answer;

   my(@all, @any);

   if($a =~ /[*\[\]\|]/) {

      if($a =~ /\*/) {
         @all = $a =~ /\*(.*?)\*/g;
      } elsif ($a =~ /\|\|/) {
         @any = split /\|\|/, $a;
      } else {
         #[]-thingy for numerical answers.
         $all[0] = $a;
      }

   } else {
      $all[0] = $a;
   }

   return (\@all, \@any);

}

=item format_answer()

Returns the answer without metacharacters and properly formatted.

=cut

sub format_answer
{
   my $self = shift;

   my $a = $self->answer;

   if($a =~ /\|\|/) {
      my @a = split /\|\|/, $a; # $a = "foo||bar||baz"
      $a = sprintf("%s (%s)", shift @a, (join ", ", @a)); # $a = "foo (bar, baz)"
   } elsif($a =~ /\[(\d+?)\|\d+?\]/) {
      $a = $1;
   } else {
   }
         
   $a =~ s/[\|\*\[\]]//g; #remove meta-characters.

   return $a;
}


=item check_answer()

Takes an attempted answer as argument and returns true if it is valid.

=cut

sub check_answer
{
   my $self = shift;
   my $attempt = shift;

   my ($all, $any) = $self->regexes;

   my $x = 0;
   #Regexes that must match:
   foreach my $r (@$all) {
      if($r =~ /\[.*\]/) {
         #Numerical answer: (42 shouldn't match 442, so special rules apply)
         #If it's numerical, @all only contains one element.

         if($r =~ /\|/) {
            #... with bounds (e.g. [6000|50], to allow approximate answers, i.e. from 5950 to 6050)
            $r =~ /\[(.*)\]/;
            my ($x, $y) = split /\|/, $1, 2;
	    $attempt =~ s/\D//g;
	    return 0 unless($attempt);
            if($attempt <= $x + $y && $attempt >= $x - $y) {
               return 1;
            } else {
               return 0;
            }
         } else {
	    #without bounds.
            $r =~ s/[\[\]]//g;
            $attempt =~ s/\D//g;

            if($attempt == $r) {
               return 1;
            } else {
               return 0;
            }
         }
      } else {
         return 0 unless $attempt =~ /$r/i;
	 $attempt =~ s/$r//i; #Remove it so it can't match again.
         $x = 1;
      }
   } #foreach regex

   return 1 if $x;

   #Ok
   foreach my $r (@$any) {
      if( $attempt =~ /$r/i ) {
	 $self->{_answer} = $r;
	 return 1;
      }
   }

   return 0;
}



=item Query ({ ..., Select => "count(*)" })

Returns a SQL-query. See C<GetRandom> for options.

B<Select> is what is selected (default: "*"), as in "SELECT foo FROM ...". For example, "count(*)" might be useful to get the number of questions matching the rule set.

The query ends in the "WHERE"-section, so "ORDER BY"-s, etc. might be added.

=cut

sub Query
{
   my $self = shift;
   my $args = shift;

   $args->{Select} ||= '*';
   my $query = sprintf "SELECT %s FROM questions ", $args->{Select};

   my @NotIn = @{ $args->{CategoryNotIn} } if($args->{CategoryNotIn} and ref($args->{CategoryNotIn}) eq 'ARRAY');
   my @In = @{ $args->{CategoryIn} } if($args->{CategoryIn} and ref($args->{CategoryIn}) eq 'ARRAY');

   for(@NotIn) { die "Invalid category id '$_'" if $_ =~ /\D/; }
   for(@In) { die "Invalid category id '$_'" if $_ =~ /\D/; }

   my $where = 0;

   if(@NotIn) {
     $query .= "WHERE category NOT IN " . join("AND category NOT IN ", map { $_ = "(SELECT id FROM walk_tree($_)) "; } @NotIn);
     $where = 1;
   }

   if(@In) {
      if($where) {
	 $query .= "AND ";
      } else {
	 $query .= "WHERE ";
         $where = 1;
      }
	 $query .= "category IN " . join("OR category IN ", map { $_ = "(SELECT id FROM walk_tree($_)) "; } @In);
   }

   if($args->{DiffLT} && $args->{DiffLT} !~ /\D/) {
      if($where) {
         $query .= "AND ";
      } else {
         $where = 1;
         $query .= "WHERE ";
      }

      $query .= sprintf("question_score(asked, answered) < %f ", $args->{DiffLT});
   }

   if($args->{DiffGT} && $args->{DiffGT} !~ /\D/) {
      if($where) {
         $query .= "AND ";
      } else {
         $where = 1;
         $query .= "WHERE ";
      }

      $query .= sprintf("question_score(asked, answered) > %f ", $args->{DiffGT});
   }

   if($args->{round} and $args->{round} !~ /\D/) {
      if($where) {
         $query .= "AND ";
      } else {
         $where = 1;
         $query .= "WHERE ";
      }

      $query .= sprintf("id NOT IN (SELECT question FROM round_questions WHERE round=%i)", $args->{round});
      #$query .= sprintf("tagged ='t' AND id NOT IN (SELECT question FROM round_questions WHERE round=%i)", $args->{round});
   }

   return $query;
}

=item save

Writes the changes to the database. The following options may be modified:

B<question>: The question. :-)

B<answer>: The, uh, answer. :-)

B<author>: The user ID of the user that has written the question. (Integer)

B<category>: The category ID the question belongs to. (Integer)

B<active>: Boolean. The question is only asked if it's active.

B<tagged>: Boolean. The question is marked for review.

=cut

sub save
{
   my $self = shift;

   my %row = map { $_ => $self->{$_} } qw(author answer question category active tagged);
   $row{id} = $self->id;

   #Convert bools:
   $row{active} ? $row{active} = 't' : $row{active} = 'f';
   $row{tagged} ? $row{tagged} = 't' : $row{tagged} = 'f';

   dbi_save_from_hash($dbh, 'questions', 'id', \%row); 

}

=item id

Returns the question's id. (Read only)

=cut

sub id { return shift->{id}; }

=item score

Returns the question's score/difficulty.

=cut

sub score
{
    my $self = shift;

    return dbi_exec_for_scalar($dbh, "SELECT question_score(asked, answered) AS score FROM questions WHERE id=?", $self->id);
}

=item author

Returns a C<Quiz::User>-object for the question's author.

=cut

sub author
{
   my $self = shift;

   $self->{Author} = Quiz::User->GetByID($self->{author}) unless $self->{Author};

   return $self->{Author};
}

=item category

Returns a C<Quiz::Category>-object for the question's category.

=cut

sub category
{
   my $self = shift;

   $self->{Category} = Quiz::Category->GetByID($self->{category}) unless $self->{Category};

   return $self->{Category};
}

=item question

The question text.

=item answer

The answer to the question.

=item asked

The number of times the question has been asked.

Read only.

=item answered

The number of times the question has been answered correctly.

Read only.

=item last_access

Last time the question was asked.

Read only.

=back

=cut

sub AUTOLOAD {
    my $self = shift;
    my $attr = our $AUTOLOAD;
    $attr =~ s/.*:://;
    return unless $attr =~ /[^A-Z]/;  # skip DESTROY and all-cap methods

    my %ok_field = map { $_ => 1 } qw(question answer asked answered last_accessed);
    die "invalid attribute method: ->$attr( )" unless $ok_field{$attr};

    $self->{lc $attr} = shift if @_;
    return $self->{lc $attr};
}

=head1 SEE ALSO

L<Quiz::Round>
L<Quiz::User>
L<Quiz::Category>

=head1 AUTHOR

Alex Brasetvik 

=cut

1;
