DROP TABLE users CASCADE;

CREATE TABLE users (
	id SERIAL PRIMARY KEY,
	username VARCHAR(9) UNIQUE NOT NULL,
	hostmask VARCHAR(255),
	registered DATE DEFAULT NOW(),
	last_login TIMESTAMP(0),
	flags VARCHAR(255),
	password CHAR(32) NOT NULL, -- md5 sum
	active BOOLEAN DEFAULT 't' NOT NULL -- users shouldn't be deleted, but deactivated.
);

CREATE INDEX user_hostmask ON users(hostmask);
CREATE INDEX user_active ON users(hostmask);

-- A hostmask can only be logged in as one user at a time:
CREATE OR REPLACE FUNCTION hostmask() RETURNS trigger AS '
	BEGIN
		UPDATE users SET hostmask = null WHERE hostmask = NEW.hostmask AND id != NEW.id;
		NEW.last_login = now();
		RETURN NEW;
	END;
' LANGUAGE plpgsql;

CREATE TRIGGER user_hostmask_trigger BEFORE INSERT OR UPDATE ON users
	FOR EACH ROW
		EXECUTE PROCEDURE hostmask();

/*
DROP TABLE user_meta CASCADE;

CREATE TABLE user_meta (
	user_id INTEGER NOT NULL REFERENCES users(id)
		ON UPDATE CASCADE
		ON DELETE CASCADE,
	namespace VARCHAR(455) NOT NULL,
	key VARCHAR(455) NOT NULL,
	value VARCHAR(455) NOT NULL,
	PRIMARY KEY (user_id, namespace, key)
);

DROP TABLE user_stats CASCADE;

CREATE TABLE user_stats (
	user_id INTEGER NOT NULL REFERENCES users(id)
		ON UPDATE CASCADE
		ON DELETE CASCADE,
	namespace VARCHAR(455) NOT NULL,
	key VARCHAR(455) NOT NULL,
	value VARCHAR(455) NOT NULL,
	time TIMESTAMP(0) NOT NULL DEFAULT NOW(),
	PRIMARY KEY (user_id, namespace, key, time)
);

*/

DROP TABLE categories CASCADE;

CREATE TABLE categories (
    id SERIAL PRIMARY KEY,
    parent INTEGER REFERENCES categories(id)
    	ON UPDATE CASCADE
	ON DELETE CASCADE
	CONSTRAINT parent_ok CHECK (parent <> id),
    name VARCHAR(255) NOT NULL,
    UNIQUE(parent, name)
);

CREATE INDEX category_name ON categories(name);
CREATE INDEX category_parent ON categories(parent);

DROP FUNCTION walk_tree(int);
DROP FUNCTION walk_tree(int, int);
DROP TYPE category_rt CASCADE;

CREATE TYPE category_rt AS (
  level INTEGER,
  id INTEGER,
  parent INTEGER,
  name VARCHAR
);

-- Given a category id, this procedure traverses the category tree, and returns
-- every child.

CREATE FUNCTION walk_tree(int, int) RETURNS SETOF category_rt AS '
DECLARE
  row categories%rowtype;
  retrow category_rt%rowtype;
BEGIN
  FOR row IN EXECUTE ''SELECT id,parent,name FROM categories WHERE id = '' || quote_literal($1) LOOP
    retrow.level := $2;
    retrow.id := row.id;
    retrow.parent := row.parent;
    retrow.name := row.name;
    RETURN NEXT retrow;
  END LOOP;
  FOR row IN EXECUTE ''SELECT id,parent,name FROM categories WHERE parent = '' || quote_literal($1) LOOP
    FOR retrow IN SELECT * FROM walk_tree(row.id,$2+1) LOOP
      RETURN NEXT retrow;
    END LOOP;
  END LOOP;
  RETURN;
END;
' LANGUAGE 'plpgsql';

CREATE FUNCTION walk_tree(int) RETURNS SETOF category_rt AS '
	SELECT * FROM walk_tree($1, 0)
' LANGUAGE 'sql';
--'

CREATE OR REPLACE FUNCTION breadcrumb(int, varchar) RETURNS VARCHAR AS '
DECLARE
	breadcrumb varchar;
	p record;
BEGIN
	SELECT INTO p * FROM categories WHERE id = $1;
	breadcrumb := p.name;

	WHILE p.parent IS NOT NULL LOOP
		SELECT INTO p * FROM categories WHERE id = p.parent;
		breadcrumb := p.name || $2 || breadcrumb;
	END LOOP;

	RETURN breadcrumb;
END;
' LANGUAGE 'plpgsql';
--'
/*
INSERT INTO categories (parent, name) VALUES (NULL,'Vitenskap');
INSERT INTO categories (parent, name) VALUES (NULL,'Geografi');
INSERT INTO categories (parent, name) VALUES (NULL,'Musikk');
INSERT INTO categories (parent, name) VALUES (1,'Matte');
INSERT INTO categories (parent, name) VALUES (1,'Kjemi');
INSERT INTO categories (parent, name) VALUES (2,'Europa');
INSERT INTO categories (parent, name) VALUES (2,'Asia');
INSERT INTO categories (parent, name) VALUES (3,'Pop');
INSERT INTO categories (parent, name) VALUES (4,'Integraler');
INSERT INTO categories (parent, name) VALUES (4,'Sannsynlighet');
INSERT INTO categories (parent, name) VALUES (NULL,'Sport');
INSERT INTO categories (parent, name) VALUES (11,'Fotball');
INSERT INTO categories (parent, name) VALUES (10,'Bayes');
*/
-- Once added, a question should never be deleted, as lots of references are made to them.
DROP TABLE questions CASCADE; 

CREATE TABLE questions (
	id SERIAL PRIMARY KEY
		CONSTRAINT question_id_ok CHECK (id > -1),
	author INTEGER REFERENCES users(id)
		ON UPDATE CASCADE
		ON DELETE SET NULL,
	active BOOLEAN DEFAULT 't', -- Don't delete questions - deactivate them.
	tagged BOOLEAN DEFAULT 'f', -- Questions that suck should be marked for review.
	category INTEGER REFERENCES categories(id)
		ON UPDATE CASCADE
		ON DELETE SET NULL, -- Make it an orphan rather than deleting it.
	last_access DATE,
	question VARCHAR(455) NOT NULL,
	answer VARCHAR(455) NOT NULL,
	asked INTEGER DEFAULT 0 -- Number of times the question has been asked.
		CONSTRAINT asked_ok CHECK (asked > -1 AND asked >= answered),
	answered INTEGER DEFAULT 0 -- Number of times the question has been answered correctly.
		CONSTRAINT answered_ok CHECK (answered > -1 AND answered <= asked),
	UNIQUE (category, question)
);

CREATE OR REPLACE FUNCTION question_access() RETURNS trigger AS '
	BEGIN
		NEW.last_access = now();
		RETURN NEW;
	END;
' LANGUAGE plpgsql;

--'

CREATE TRIGGER question_access_trigger BEFORE UPDATE ON questions
	FOR EACH ROW
		EXECUTE PROCEDURE question_access();
/*
INSERT INTO questions(author, category, question, answer) VALUES (1, 6, 'Hva er hovedstaden i Norge?','Oslo');
INSERT INTO questions(author, category, question, answer) VALUES (1, 6, 'Hva er hovedstaden i Sverige?','Stockholm');
INSERT INTO questions(author, category, question, answer) VALUES (1, 6, 'Hva er hovedstaden i Danmark?','København');
INSERT INTO questions(author, category, question, answer) VALUES (1, 6, 'Hva er hovedstaden i Finnland?','Helsinki');
INSERT INTO questions(author, category, question, answer) VALUES (1, 6, 'Hva er hovedstaden i Island?','Reykjavik');
INSERT INTO questions(author, category, question, answer) VALUES (1, 6, 'Hva er hovedstaden i England?','London');
INSERT INTO questions(author, category, question, answer) VALUES (1, 6, 'Hva er hovedstaden i Tyskland?','Berlin');
INSERT INTO questions(author, category, question, answer) VALUES (1, 6, 'Hva er hovedstaden i Østerrike?','Wien');
INSERT INTO questions(author, category, question, answer) VALUES (1, 6, 'Hva er hovedstaden i Frankrike?','Paris');
INSERT INTO questions(author, category, question, answer) VALUES (1, 7, 'Hva er hovedstaden i Kina?','Beijing');
INSERT INTO questions(author, category, question, answer) VALUES (1, 10, 'Hva er sannsynligheten for to seksere på rad?','1/36');
*/

DROP TABLE question_reports CASCADE;

CREATE TABLE question_reports (
	id SERIAL,
	question INTEGER NOT NULL REFERENCES questions(id)
		ON UPDATE CASCADE
		ON DELETE CASCADE,
	author INTEGER NOT NULL REFERENCES users(id)
		ON UPDATE CASCADE
		ON DELETE CASCADE,
	comment VARCHAR(455)
);

CREATE OR REPLACE FUNCTION question_score(int,int) RETURNS real AS '
        DECLARE
                asked ALIAS FOR $1;
                answered ALIAS FOR $2;
                difficulty real DEFAULT 0.5;
                n real;
        BEGIN
                IF asked > 0 THEN
                        difficulty := float4(answered) / float4(asked);
                END IF;
                
                n := sqrt(float4(asked)/10);
		n := exp(1 - 2*difficulty)^(n);

		IF n > 10 THEN
			RETURN 10;
		END IF;
                
		RETURN round(n::numeric,2);
        END;
' LANGUAGE 'plpgsql';
--'

CREATE OR REPLACE FUNCTION reduced_question_score(int,int) RETURNS real AS '
        DECLARE
		qid ALIAS FOR $1;
		uid ALIAS FOR $2;
		question questions%rowtype;
		user_answered_count int;
        BEGIN
		SELECT INTO question * FROM questions WHERE id=qid;
		IF NOT FOUND THEN
			RAISE EXCEPTION ''No such question.'';
		END IF;

		SELECT INTO user_answered_count count(*) FROM round_questions WHERE question=qid AND winner=uid;
		RETURN round((question_score(question.asked, question.answered) * (exp(-sqrt(user_answered_count))^(0.25)))::numeric, 2);
        END;
' LANGUAGE 'plpgsql';

DROP TABLE rounds CASCADE;

CREATE TABLE rounds (
	id SERIAL PRIMARY KEY 
		CONSTRAINT round_id_ok CHECK (id > -1),
	started TIMESTAMP(0) NOT NULL DEFAULT now(),
	ended TIMESTAMP(0),
	mode INTEGER DEFAULT 0, -- e.g. teamplay, etc.
	winner INTEGER REFERENCES users(id)
		ON UPDATE CASCADE
		ON DELETE SET NULL,
	winner_team INTEGER
);

CREATE INDEX round_started ON rounds(started);
CREATE INDEX round_winner ON rounds(winner);


DROP TABLE round_meta CASCADE;

CREATE TABLE round_meta (
	round INTEGER NOT NULL REFERENCES rounds(id)
		ON UPDATE CASCADE
		ON DELETE CASCADE,
	key VARCHAR(455) NOT NULL,
	value VARCHAR(455) NOT NULL,
	PRIMARY KEY (round, key)
);

DROP TABLE round_questions CASCADE;

CREATE TABLE round_questions (
	round INTEGER NOT NULL REFERENCES rounds(id)
		ON UPDATE CASCADE
		ON DELETE CASCADE,
	question INTEGER NOT NULL REFERENCES questions(id)
		ON UPDATE CASCADE
		ON DELETE CASCADE,
	winner INTEGER REFERENCES users(id)
		ON UPDATE CASCADE
		ON DELETE SET NULL,
	answers INTEGER
		CONSTRAINT answers_ok CHECK (answers > -1),
	-- I'm not sure how it scales calculating the score with question_score() all the time, so save the score
	-- until it's tested. :)
	score FLOAT,
	date TIMESTAMP(0) NOT NULL DEFAULT now(),
	PRIMARY KEY(round, question)
);

CREATE INDEX round_question_winners ON round_questions(winner);

DROP TABLE round_participants CASCADE;
CREATE TABLE round_participants (
	round INTEGER REFERENCES rounds(id)
		ON UPDATE CASCADE
		ON DELETE CASCADE,
	user_id INTEGER REFERENCES users(id)
		ON UPDATE CASCADE
		ON DELETE CASCADE,
	attempts INTEGER DEFAULT 0
		CONSTRAINT attempts_ok CHECK (attempts >= wins),
	wins INTEGER DEFAULT 0
		CONSTRAINT wins_ok CHECK (attempts >= wins),
	PRIMARY KEY (round, user_id)
);

CREATE INDEX round_user_ids ON round_participants(user_id);
CREATE INDEX round_round_ids ON round_participants(round);

DROP TABLE round_teams CASCADE;
CREATE TABLE round_teams (
	round INTEGER NOT NULL REFERENCES rounds(id)
		ON UPDATE CASCADE
		ON DELETE CASCADE,
	user_id INTEGER REFERENCES users(id)
		ON UPDATE CASCADE
		ON DELETE SET NULL,
	team INTEGER,
	PRIMARY KEY (round, user_id)
);

CREATE INDEX round_team_users ON round_teams(user_id);

CREATE OR REPLACE FUNCTION round_team_okay() RETURNS trigger AS '
	BEGIN
		PERFORM id FROM rounds WHERE id = NEW.round AND mode=1;

		IF NOT FOUND THEN
			RAISE EXCEPTION ''Round % is not a team game'', NEW.round;
		END IF;
		RETURN NEW;
	END;
' LANGUAGE plpgsql;
--'
CREATE TRIGGER round_team_trigger BEFORE INSERT OR UPDATE ON round_teams
	FOR EACH ROW
		EXECUTE PROCEDURE round_team_okay();

-- Round 1. UID 1 wins. (round win != question win)
/*
INSERT INTO rounds (ended, winner) VALUES (now(), 1);

INSERT INTO round_questions (round, question, winner, answers) VALUES (1, 1, 1, 3);
INSERT INTO round_questions (round, question, winner, answers) VALUES (1, 2, null, 3);
INSERT INTO round_questions (round, question, winner, answers) VALUES (1, 3, 1, 3);
INSERT INTO round_questions (round, question, winner, answers) VALUES (1, 4, null, 3);
INSERT INTO round_questions (round, question, winner, answers) VALUES (1, 5, 2, 3);
INSERT INTO round_questions (round, question, winner, answers) VALUES (1, 6, 3, 3);

INSERT INTO round_participants (round, user_id, attempts, wins) VALUES (1, 1, 6, 3);
INSERT INTO round_participants (round, user_id, attempts, wins) VALUES (1, 2, 6, 2);
INSERT INTO round_participants (round, user_id, attempts, wins) VALUES (1, 3, 6, 1);

-- Round 2. "Team play".
INSERT INTO rounds (ended, mode, winner_team) VALUES (now(), 1, 2);

INSERT INTO round_teams (round, user_id, team) VALUES (2, 1, 1); -- Three one man armys.
INSERT INTO round_teams (round, user_id, team) VALUES (2, 2, 2);
INSERT INTO round_teams (round, user_id, team) VALUES (2, 3, 3);

INSERT INTO round_questions (round, question, winner, answers) VALUES (2, 1, 1, 3);
INSERT INTO round_questions (round, question, winner, answers) VALUES (2, 2, null, 3);
INSERT INTO round_questions (round, question, winner, answers) VALUES (2, 3, 2, 3);
INSERT INTO round_questions (round, question, winner, answers) VALUES (2, 4, 2, 3);
INSERT INTO round_questions (round, question, winner, answers) VALUES (2, 5, 3, 3);
INSERT INTO round_questions (round, question, winner, answers) VALUES (2, 1, null, 3);

INSERT INTO round_participants (round, user_id, attempts, wins) VALUES (2, 1, 6, 1);
INSERT INTO round_participants (round, user_id, attempts, wins) VALUES (2, 2, 6, 3);
INSERT INTO round_participants (round, user_id, attempts, wins) VALUES (2, 3, 6, 2);
*/
CREATE OR REPLACE VIEW scores AS
	SELECT	id, username,
		coalesce((SELECT count(winner) FROM rounds WHERE winner = users.id)  -- Freelance wins.
		+ (SELECT count(rounds.id) FROM rounds, round_teams
			WHERE rounds.mode=1 AND rounds.id=round_teams.round
			AND rounds.winner_team=round_teams.team AND round_teams.user_id=users.id
		),0)   -- Team wins.
		AS wins, -- Number of times he has won.
/*
-- This will enable "floating" score:
		coalesce((SELECT sum(iscore) FROM ( -- sum ... 
					SELECT question_score(questions.asked,questions.answered) AS iscore FROM round_questions, questions -- ... the score for
					WHERE questions.active = 't' AND		  -- ... the active and
						round_questions.question = questions.id AND
						round_questions.winner=users.id		   -- ... winning questions.
				) AS temp
		),0) AS total_score,
*/
		coalesce((SELECT sum(score) FROM round_questions WHERE winner=users.id), 0)::real AS total_score,
		coalesce((SELECT sum(wins) FROM round_participants WHERE user_id=users.id),0) AS answered,
		coalesce((SELECT sum(attempts) FROM round_participants WHERE user_id=users.id),1) AS attempts
		FROM users WHERE users.active = 't';

DROP SEQUENCE score_rank;
CREATE SEQUENCE score_rank MINVALUE 1;

DROP TABLE score_cache;
CREATE TABLE score_cache (
	id INTEGER PRIMARY KEY REFERENCES users(id)
		ON UPDATE CASCADE
		ON DELETE CASCADE,
	wins INTEGER DEFAULT 0,
	total_score REAL DEFAULT 0,
	answered INTEGER DEFAULT 0,
	attempts INTEGER DEFAULT 0
);

CREATE INDEX score_index ON score_cache(total_score);

CREATE OR REPLACE FUNCTION rebuild_scores() RETURNS INTEGER AS '
BEGIN
	DELETE FROM score_cache;
	INSERT INTO score_cache (id, wins, total_score, answered, attempts) SELECT id, wins, total_score, answered, attempts FROM scores;
	UPDATE score_cache SET attempts=1 WHERE attempts=0;
	RETURN 1;
END;
' LANGUAGE plpgsql;

-- Don't screw with old data.
CREATE OR REPLACE FUNCTION round_update() RETURNS trigger AS '
	BEGIN
		IF OLD.ended IS NOT NULL
		THEN 
 			RETURN OLD;
		END IF;

		IF new.winner IS NOT NULL
		THEN
			UPDATE score_cache SET wins = wins + 1 WHERE id=new.winner;
		END IF;

		RETURN NEW;
	END;
' LANGUAGE plpgsql;
--'

CREATE TRIGGER round_update_ended_trigger BEFORE UPDATE ON rounds
	FOR EACH ROW
		EXECUTE PROCEDURE round_update();



-- Ensure the user has a row in the score_cache-table.
CREATE OR REPLACE FUNCTION round_new_participant() RETURNS trigger AS '
	DECLARE
		score_row score_cache%ROWTYPE;
	BEGIN
		SELECT INTO score_row * FROM score_cache WHERE id=new.user_id;
		IF NOT FOUND
		THEN
	           INSERT INTO score_cache (id, wins, total_score, answered, attempts)
		   	SELECT id, wins, total_score, answered, attempts FROM scores WHERE id=new.user_id;
	        END IF;

		RETURN NEW;
	END;
' LANGUAGE plpgsql;
--'

CREATE TRIGGER round_new_participant AFTER INSERT ON round_participants
	FOR EACH ROW
		EXECUTE PROCEDURE round_new_participant();

-- Update scores:
CREATE OR REPLACE FUNCTION user_update_attempts() RETURNS trigger AS '
	BEGIN
		IF new.attempts > old.attempts
		THEN
		   UPDATE score_cache SET attempts = attempts + (new.attempts - old.attempts) WHERE id=new.user_id;
		END IF;

		IF new.wins > old.wins -- Correct answers, not round wins.
		THEN
		   UPDATE score_cache SET answered = answered + (new.wins - old.wins) WHERE id=new.user_id;
		END IF;

		RETURN NEW;
	END;
' LANGUAGE plpgsql;
--'

CREATE TRIGGER user_update_attempts AFTER UPDATE ON round_participants
	FOR EACH ROW
		EXECUTE PROCEDURE user_update_attempts();

CREATE OR REPLACE FUNCTION round_update_stats() RETURNS trigger AS '
	BEGIN
		IF new.winner IS NOT NULL THEN
			SELECT INTO new.score reduced_question_score(new.question, new.winner) FROM questions WHERE id = new.question;
			UPDATE score_cache SET total_score = total_score + new.score WHERE id=new.winner;
			UPDATE questions SET asked = asked + 1, answered = answered + 1 WHERE id = new.question;
		ELSE 
		        UPDATE questions SET asked = asked + 1 WHERE id = new.question;
		END IF;

		RETURN NEW;
	END;
' LANGUAGE plpgsql;
--'
CREATE TRIGGER round_update_trigger BEFORE INSERT ON round_questions
	FOR EACH ROW
		EXECUTE PROCEDURE round_update_stats();


DROP TABLE rank_cache CASCADE;

CREATE TABLE rank_cache (
	rank INTEGER NOT NULL DEFAULT nextval('score_rank'),
	id INTEGER NOT NULL REFERENCES users(id)
		ON UPDATE CASCADE
		ON DELETE CASCADE,
	PRIMARY KEY(id)
);

DROP TABLE rank_history;
CREATE TABLE rank_history (
	id INTEGER NOT NULL REFERENCES users(id)
		ON UPDATE CASCADE
		ON DELETE CASCADE,
	wins INTEGER,
	total_score REAL,
	answered INTEGER,
	attempts INTEGER,
	rank INTEGER,
	date TIMESTAMP(0) NOT NULL DEFAULT NOW(),
	PRIMARY KEY(id, date)
);


CREATE OR REPLACE FUNCTION rebuild_ranks() RETURNS INTEGER AS '
BEGIN
	INSERT INTO rank_history (id, wins, total_score, answered, attempts, rank) SELECT score_cache.id, wins, total_score, answered, attempts, rank FROM score_cache, rank_cache WHERE score_cache.id=rank_cache.id AND total_score IS NOT NULL AND score_cache.id IN (SELECT user_id FROM round_participants WHERE round >=(SELECT MAX(id)-1 FROM rounds));
	DELETE FROM rank_cache;
	PERFORM setval(''score_rank'',1,false); -- Reset rank counter.
	INSERT INTO rank_cache (id) SELECT id FROM score_cache ORDER BY total_score IS NOT NULL DESC, total_score DESC, wins IS NOT NULL DESC, wins DESC;
	RETURN 1;
END;
' LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION round_update_ranks() RETURNS trigger AS '
	BEGIN
		IF NEW.ended IS NOT NULL
		THEN 
 			PERFORM rebuild_ranks();
		END IF;
		RETURN NEW;
	END;
' LANGUAGE plpgsql;
--'

-- XXX
CREATE TRIGGER round_updated_ended_trigger AFTER UPDATE ON rounds
	FOR EACH ROW
		EXECUTE PROCEDURE round_update_ranks();

-- if the previous round wasn't ended sanely, rebuild ranks.

CREATE OR REPLACE FUNCTION round_ensure_fresh_ranks() RETURNS trigger AS '
	DECLARE
		pround rounds%ROWTYPE;
	BEGIN
		SELECT INTO pround * FROM rounds WHERE id = NEW.id - 1;
		IF pround.ended IS NULL
		THEN 
 			PERFORM rebuild_ranks();
		END IF;
		RETURN NEW;
	END;
' LANGUAGE plpgsql;
--'

-- XXX
CREATE TRIGGER round_inserted_trigger AFTER INSERT ON rounds
	FOR EACH ROW
		EXECUTE PROCEDURE round_ensure_fresh_ranks();

CREATE OR REPLACE VIEW userinfo AS 
	SELECT
		users.id, users.username, users.hostmask, users.registered, users.flags, users.password,
		rank_cache.rank, score_cache.wins, score_cache.total_score AS score, score_cache.answered, score_cache.attempts,
		100 * float4(score_cache.answered)/float4(score_cache.attempts) AS ratio
	FROM users, rank_cache, score_cache
	WHERE users.id = rank_cache.id AND users.id = score_cache.id
		AND users.active = 't';

DROP TYPE rank_rt CASCADE;
CREATE TYPE rank_rt AS (
  rank INTEGER,
  username VARCHAR,
  wins INTEGER,
  score REAL,
  ratio double precision
);

CREATE OR REPLACE FUNCTION top(integer) RETURNS SETOF rank_rt AS '
		SELECT rank, username, wins, score, ratio
		FROM userinfo
		ORDER BY rank ASC, score DESC, ratio DESC, attempts DESC
		LIMIT $1;
' LANGUAGE SQL;

SELECT * FROM top(2);

/* Example:
SELECT * from top(2);
 rank | username | wins |  score  |      ratio
------+--------+------+---------+------------------
    1 | fnupi  |    1 | 2.89498 | 41.6666666666667
    2 | alex   |    1 | 2.85124 | 33.3333333333333
*/

CREATE OR REPLACE FUNCTION low(integer) RETURNS SETOF rank_rt AS '
		SELECT rank, username, wins, score, ratio
		FROM userinfo
		WHERE score IS NOT NULL
		ORDER BY rank DESC, score ASC, ratio ASC, attempts ASC
		LIMIT $1;
' LANGUAGE SQL;


